@@ -1,13 +1,19 @@
 # robot imports
+
 from robot.api.deco import library, keyword
 from robot.api.logger import librarylogger as logger
 
+# requests import
+from requests import HTTPError
+import  requests
+from requests_toolbelt.multipart.encoder import MultipartEncoder
+
 # python imports
 import os
 from typing import List, Dict, Any
 import time
 
-from generic_camunda_client import ApiException, DeploymentWithDefinitionsDto, LockedExternalTaskDto, \
+from generic_camunda_client import ApiException, DeploymentWithDefinitionsDto, DeploymentDto, LockedExternalTaskDto, \
     VariableValueDto, FetchExternalTasksDto, FetchExternalTaskTopicDto, ProcessDefinitionApi, \
     ProcessInstanceWithVariablesDto, StartProcessInstanceDto, ProcessInstanceModificationInstructionDto,\
     ProcessInstanceApi, ProcessInstanceDto
@@ -22,6 +28,17 @@ class CamundaLibrary:
     """
     Library for Camunda integration in Robot Framework
 
+    = Installation =
+
+    == Camunda ==
+
+    Easiest to run Camunda in docker:
+    | docker run -d --name camunda -p 8080:8080 camunda/camunda-bpm-platform:run-latest
+
+    == CamundaLibrary ==
+    You can use pip for installing CamundaLibrary:
+    | pip install robotframework-camunda
+
     = Usage =
 
     The library provides convenience keywords for accessing Camunda via REST API. You may deploy models,
@@ -49,8 +66,8 @@ class CamundaLibrary:
     You may achieve a kind of subscription by providing the ``asyncResponseTimeout`` with the `Fetch workload`
     keyword in order to achieve [https://docs.camunda.org/manual/7.14/user-guide/process-engine/external-tasks/#long-polling-to-fetch-and-lock-external-tasks|Long Polling].
 
-    | ${variables} | fetch workload | my_topic | asyncResponseTimeout=60000 |
-    | log | Waited at most 1 hour before this log statement got executed |
+    | ${variables} | fetch workload | my_topic | async_response_timeout=60000 |
+    | log | Waited at most 1 minute before this log statement got executed |
 
     = Missing Keywords =
 
@@ -67,7 +84,7 @@ class CamundaLibrary:
 
     EMPTY_STRING = ""
     KNOWN_TOPICS: Dict[str,Dict[str, Any]] = {}
-    FETCH_RESPONSE: LockedExternalTaskDto
+    FETCH_RESPONSE: LockedExternalTaskDto = None
 
     def __init__(self, camunda_engine_url: str = 'http://localhost:8080'):
         self._shared_resources = CamundaResources()
@@ -85,26 +102,38 @@ class CamundaLibrary:
         self._shared_resources.camunda_url = f'{url}/engine-rest'
 
     @keyword(name='Deploy model from file', tags=['deployment'])
-    def deploy_bpmn(self, path_bpmn_file: str):
-        """Uploads a camunda model to camunda that is provided as path.
+    def deploy_model_from_file(self, path_to_model):
+        """*DEPRECATED*
+
+        Use `fetch workload`
+        """
+        logger.warn('Keyword "Fetch and Lock workloads" is deprecated. Use "Fetch workload" instead.')
+        return self.deploy(path_to_model)
+
+    @keyword(name='Deploy', tags=['deployment'])
+    def deploy(self, *args):
+        """Creates a deployment from all given files and uploads them to camunda.
 
         Return response from camunda rest api as dictionary. Further documentation: https://docs.camunda.org/manual/7.14/reference/rest/deployment/post-deployment/
 
         By default, this keyword only deploys changed models and filters duplicates. Deployment name is the filename of
-        the model.
+        the first file.
 
         Example:
-            | ${path_to_bpm_file} | *Set Variable* | _../bpmn/my_model.bpm_ |
-            | ${response} | *Deploy model from file* | _${path_to_bpm_file}_ |
+            | ${response} | *Deploy model from file* | _../bpmn/my_model.bpnm_ | _../forms/my_forms.html_ |
         """
-        if not path_bpmn_file:
+        if not args:
             raise ValueError('Failed deploying model, because no file provided.')
 
-        filename = os.path.basename(path_bpmn_file)
+        if len(args) > 1:
+            # We have to use plain REST then when uploading more than 1 file.
+            return self.deploy_mulitple_files(*args)
+
+        filename = os.path.basename(args[0])
 
         with self._shared_resources.api_client as api_client:
             api_instance = openapi_client.DeploymentApi(api_client)
-            data = path_bpmn_file
+            data = [*args]
             deployment_name = filename
 
             try:
@@ -119,6 +148,65 @@ class CamundaLibrary:
 
         return response.to_dict()
 
+    def deploy_mulitple_files(self, *args):
+        """
+        # Due to https://jira.camunda.com/browse/CAM-13105 we cannot use generic camunda client when dealing with
+        # multiple files. We have to use plain REST then.
+        """
+
+        fields = {
+            'deployment-name': f'{os.path.basename(args[0])}',
+            'enable-duplicate-filtering': "true",
+            'deploy-changed-only': "true",
+        }
+
+        for file in args:
+            filename = os.path.basename(file)
+            fields[f'{filename}'] = (filename, open(file, 'rb'), 'application/octet-stream')
+
+        multipart_data = MultipartEncoder(
+            fields=fields
+        )
+
+        logger.debug(multipart_data.fields)
+
+        response = requests.post(f'{self._shared_resources.camunda_url}/deployment/create', data=multipart_data,
+                                 headers={'Content-Type': multipart_data.content_type})
+        json = response.json()
+        try:
+            response.raise_for_status()
+            logger.debug(json)
+        except HTTPError as e:
+            logger.error(json)
+            raise e
+
+        return json
+
+    @keyword(name='Get deployments', tags=['deployment'])
+    def get_deployments(self, deployment_id: str = None, **kwargs):
+        """
+        Retrieves all deployments that match given criteria. All parameters are available from https://docs.camunda.org/manual/latest/reference/rest/deployment/get-query/
+
+        Example:
+            | ${list_of_deployments} | get deployments | ${my_deployments_id} |
+            | ${list_of_deployments} | get deployments | id=${my_deployments_id} |
+            | ${list_of_deployments} | get deployments | after=2013-01-23T14:42:45.000+0200 |
+        """
+        if deployment_id:
+            kwargs['id'] = deployment_id
+
+        with self._shared_resources.api_client as api_client:
+            api_instance = openapi_client.DeploymentApi(api_client)
+
+            try:
+                response: List[DeploymentDto] = api_instance.get_deployments(**kwargs)
+                logger.info(f'Response from camunda:\t{response}')
+            except ApiException as e:
+                logger.error(f'Failed get deployments:\n{e}')
+                raise e
+
+        return [r.to_dict() for r in response]
+
     @keyword("Fetch and Lock workloads", tags=['task', 'deprecated'])
     def fetch_and_lock_workloads(self, topic, **kwargs) -> Dict:
         """*DEPRECATED*
@@ -129,7 +217,7 @@ class CamundaLibrary:
         return self.fetch_workload(topic, **kwargs)
 
     @keyword("Fetch Workload", tags=['task'])
-    def fetch_workload(self, topic: str, **kwargs) -> Dict:
+    def fetch_workload(self, topic: str, async_response_timeout=None, use_priority=None, **kwargs) -> Dict:
         """
         Locks and fetches workloads from camunda on a given topic. Returns a list of variable dictionary.
         Each dictionary representing 1 workload from a process instance.
@@ -167,7 +255,10 @@ class CamundaLibrary:
             if 'deserialize_values' not in kwargs:
                 kwargs['deserialize_values'] = True
             topic_dto=FetchExternalTaskTopicDto(topic_name=topic, **kwargs)
-            fetch_external_tasks_dto = FetchExternalTasksDto(worker_id=self.WORKER_ID, max_tasks=1, topics=[topic_dto])
+            fetch_external_tasks_dto = FetchExternalTasksDto(worker_id=self.WORKER_ID, max_tasks=1,
+                                                             async_response_timeout=async_response_timeout,
+                                                             use_priority=use_priority,
+                                                             topics=[topic_dto])
 
             try:
                 api_response = api_instance.fetch_and_lock(fetch_external_tasks_dto=fetch_external_tasks_dto)
@@ -228,7 +319,11 @@ class CamundaLibrary:
         """
         if self.FETCH_RESPONSE:
             return self.FETCH_RESPONSE.to_dict()
-        return self.EMPTY_STRING
+        return self.FETCH_RESPONSE
+
+    @keyword("Drop fetch response")
+    def drop_fetch_response(self):
+        self.FETCH_RESPONSE = {}
 
     @keyword("Complete task", tags=['task'])
     def complete(self, result_set: Dict[str, Any] = None, files: Dict = None):
@@ -269,7 +364,7 @@ class CamundaLibrary:
                 try:
                     logger.debug(f"Sending to Camunda for completing Task:\n{complete_task_dto}")
                     api_instance.complete_external_task_resource(self.FETCH_RESPONSE.id, complete_external_task_dto=complete_task_dto)
-                    self.FETCH_RESPONSE = {}
+                    self.drop_fetch_response()
                 except ApiException as e:
                     logger.error(f"Exception when calling ExternalTaskApi->complete_external_task_resource: {e}\n")
 
@@ -301,12 +396,13 @@ class CamundaLibrary:
                 api_instance = openapi_client.ExternalTaskApi(api_client)
                 try:
                     api_instance.unlock(self.FETCH_RESPONSE.id)
+                    self.drop_fetch_response()
                 except ApiException as e:
                     logger.error(f"Exception when calling ExternalTaskApi->unlock: {e}\n")
 
     @keyword("Start process", tags=['process'])
     def start_process(self, process_key: str, variables: Dict = None, files: Dict = None,
-                      before_activity_id: str = None, after_activity_id: str = None) -> Dict:
+                      before_activity_id: str = None, after_activity_id: str = None,**kwargs) -> Dict:
         """
         Starts a new process instance from a process definition with given key.
 
@@ -322,6 +418,7 @@ class CamundaLibrary:
 
         == Examples ==
         | `start process`      | apply for job promotion    | _variables_= { 'employee' : 'John Doe', 'permission_for_application_granted' : True}   | _files_ = { 'cv' : 'documents/my_life.md'}  | _after_activity_id_ = 'Activity_ask_boss_for_persmission'   |
+        | `start process` | apply for promotion | business_key=John again |
         """
         if not process_key:
             raise ValueError('Error starting process. No process key provided.')
@@ -335,17 +432,16 @@ class CamundaLibrary:
             openapi_files = CamundaResources.convert_file_dict_to_openapi_variables(files)
             openapi_variables.update(openapi_files)
 
-            start_instructions = None
             if before_activity_id or after_activity_id:
                 instruction: ProcessInstanceModificationInstructionDto = ProcessInstanceModificationInstructionDto(
                     type='startBeforeActivity' if before_activity_id else 'startAfterActivity',
-                    activity_id=before_activity_id if before_activity_id else after_activity_id
+                    activity_id=before_activity_id if before_activity_id else after_activity_id,
                 )
-                start_instructions = [instruction]
+                kwargs.update(start_instructions=[instruction])
 
             start_process_instance_dto: StartProcessInstanceDto = StartProcessInstanceDto(
                 variables=openapi_variables,
-                start_instructions=start_instructions
+                **kwargs
             )
 
             try:
